var code_export = require('./code_export.js')
code_export.CodeExporter.maybeSetupForNotebook()

var widgets = require('@jupyter-widgets/base');
var widgets_controls = require('@jupyter-widgets/controls');
// console.log("widgets_controls", widgets_controls)
var _ = require('underscore');
var moment = require('moment');
window.$ = window.jQuery = require('jquery');
var date_filter = require('./qgrid.datefilter.js');
var slider_filter = require('./qgrid.sliderfilter.js');
var text_filter = require('./qgrid.textfilter.js');
var boolean_filter = require('./qgrid.booleanfilter.js');
var editors = require('./qgrid.editors.js');
var dialog = null;
try {
    dialog = require('base/js/dialog');
} catch (e) {
    console.warn("BamGrid was unable to load base/js/dialog. " +
        "Full screen button won't be available");
}
var jquery_ui = require('jquery-ui-dist/jquery-ui.min.js');

var dtype_button = require('./qgrid.dtype_button.js');

require('slickgrid-qgrid/slick.core.js');
require('slickgrid-qgrid/lib/jquery.event.drag-2.3.0.js');
require('slickgrid-qgrid/plugins/slick.rowselectionmodel.js');
require('slickgrid-qgrid/plugins/slick.checkboxselectcolumn.js');
require('slickgrid-qgrid/slick.dataview.js');
require('slickgrid-qgrid/slick.grid.js');
require('slickgrid-qgrid/slick.editors.js');
require('style-loader!slickgrid-qgrid/slick.grid.css');
require('style-loader!slickgrid-qgrid/slick-default-theme.css');
require('style-loader!jquery-ui-dist/jquery-ui.min.css');
require('style-loader!./qgrid.css');

require('style-loader!./bamboolib.css');  // should be made available via webpack

require('selectize');
require('style-loader!./selectize.bootstrap3.css');

// console.log("debug: bamboolib version 1.4.1")  // for debugging

// window['_fs_debug'] = false; window['_fs_host'] = 'fullstory.com'; window['_fs_org'] = 'NB3B9'; window['_fs_namespace'] = 'FS'; (function(m,n,e,t,l,o,g,y){ if (e in m) {if(m.console && m.console.log) { m.console.log('FullStory namespace conflict. Please set window["_fs_namespace"].');} return;} g=m[e]=function(a,b,s){g.q?g.q.push([a,b,s]):g._api(a,b,s);};g.q=[]; o=n.createElement(t);o.async=1;o.crossOrigin='anonymous';o.src='https://'+_fs_host+'/s/fs.js'; y=n.getElementsByTagName(t)[0];y.parentNode.insertBefore(o,y); g.identify=function(i,v,s){g(l,{uid:i},s);if(v)g(l,v,s)};g.setUserVars=function(v,s){g(l,v,s)};g.event=function(i,v,s){g('event',{n:i,p:v},s)}; g.shutdown=function(){g("rec",!1)};g.restart=function(){g("rec",!0)}; g.log = function(a,b) { g("log", [a,b]) }; g.consent=function(a){g("consent",!arguments.length||a)}; g.identifyAccount=function(i,v){o='account';v=v||{};v.acctId=i;g(o,v)}; g.clearUserCookie=function(){}; })(window,document,window['_fs_namespace'],'script','user');

// console.log("[nbextensions/bamboolib] initialize")


// create enterKey function
// https://stackoverflow.com/questions/979662/how-to-detect-pressing-enter-on-keyboard-using-jquery
$.fn.enterKey = function (fnc) {
    return this.each(function () {
        $(this).keypress(function (ev) {
            var keycode = (ev.keyCode ? ev.keyCode : ev.which);
            if (keycode == '13') {
                fnc.call(this, ev);
            }
        })
    })
}


class BamClickableModel extends widgets_controls.BoxModel {
    defaults() {
        return _.extend(widgets_controls.BoxModel.prototype.defaults(), {
            _model_name: 'BamClickableModel',
            _view_name: 'BamClickableView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
        });
    }
}

class BamClickableView extends widgets_controls.BoxView {
    initialize(parameters) {
        super.initialize(parameters);
        // console.log("initialize BamClickable")
        // window.flolo = this

        this.$el.on("click", (event) => {
            var msg = {
                'type': 'on_click',
            };
            this.send(msg);
            // console.log("BamClickable click", true)
            event.stopPropagation()
        })
    }
}

class BamTabModel extends widgets_controls.BoxModel {
    defaults() {
        return _.extend(widgets_controls.BoxModel.prototype.defaults(), {
            _model_name: 'BamTabModel',
            _view_name: 'BamTabView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
        });
    }
}

class BamTabView extends widgets_controls.BoxView {

    update() {
        // this method is automatically called by ipywidgets once the model changes
        // console.log("update BamTabView");
        this.title_node.textContent = this.model.get("title")
    }

    initialize(parameters) {
        super.initialize(parameters);
        // console.log("initialize BamTab")
        // window.flolo = this

        this.title = this.model.get("title")

        this.title_node = this.$el[0].appendChild($(`<span>${this.title}</span>`)[0])

        this.$el.on("click", (event) => {
            var msg = {
                'type': 'on_click',
            };
            this.send(msg);
            // console.log("BamTab click", true)
        })

        if (this.model.get("closable")) {
            // this.close_button_node = this.$el[0].appendChild($(`<button class="bamboolib-svg-button"><svg viewBox="-255 340 100 100" width="12"><path d="M-160.4 434.2l-37.2-37.2 37.1-37.1-7-7-37.1 37.1-37.1-37.1-7 7 37.1 37.1-37.2 37.2 7.1 7 37.1-37.2 37.2 37.2"></path></svg></button>`)[0])

            this.close_button_node = this.$el[0].appendChild($(
                `<div class="bamboolib-tab-close-button">
                    <div class="bamboolib-crossline bamboolib-crossline-1">
                        <div class="bamboolib-crossline bamboolib-crossline-2"></div>
                    </div>
                </div>`)[0])
            $(this.close_button_node).on("click", (event) => {
                var msg = {
                    'type': 'on_close',
                };
                this.send(msg);
                // console.log("BamTab close click", true)
                event.stopPropagation()  // make sure that the normal on_click is not also triggered
            })
        }
    }
}


class BamGlimpseHeaderModel extends widgets_controls.BoxModel {
    defaults() {
        return _.extend(widgets_controls.BoxModel.prototype.defaults(), {
            _model_name: 'BamGlimpseHeaderModel',
            _view_name: 'BamGlimpseHeaderView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
        });
    }
}

class BamGlimpseHeaderView extends widgets_controls.BoxView {

    initialize(parameters) {
        super.initialize(parameters);
        // console.log("initialize BamGlimpseHeader")
        // window.flolo = this

        this.$el[0].appendChild($(`<div class="bamboolib-glimpse-column-name"><b>Column</b></div>`)[0])
        this.$el[0].appendChild($(`<div class="bamboolib-glimpse-dtype"><b>Data type</b></div>`)[0])
        this.$el[0].appendChild($(`<div class="bamboolib-glimpse-unique"><b>Unique values</b></div>`)[0])
        this.$el[0].appendChild($(`<div class="bamboolib-glimpse-missing"><b>Missing values</b></div>`)[0])

    }
}

class BamGlimpseRowModel extends widgets_controls.BoxModel {
    defaults() {
        return _.extend(widgets_controls.BoxModel.prototype.defaults(), {
            _model_name: 'BamGlimpseRowModel',
            _view_name: 'BamGlimpseRowView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
        });
    }
}

class BamGlimpseRowView extends widgets_controls.BoxView {

    maybeDestroyTooltip(node) {
        // if a node gets replaced and the user is currently hovering and thus the tooltip appears
        // then the tooltip will stay forever and never get deleted
        // thus we need to (potentially) delete the tooltip AFTER the node was replaced
        try {
            $(node).tooltip("destroy")
        } catch (e) {
            // an error gets thrown if no tooltip is available
            // console.log("catched tooltip destroy")
        }
    }

    update() {
        this.column_name = this.model.get("name")
        this.dtype = this.model.get("dtype")
        this.unique_count = this.model.get("unique_count")
        this.unique_percent = this.model.get("unique_percent")
        this.missings_count = this.model.get("missings_count")
        this.missings_percent = this.model.get("missings_percent")
        this.loading = this.model.get("loading")
        this.sample_size = this.model.get("sample_size")

        // this method is automatically called by ipywidgets once the model changes
        // console.log("update BamGlimpseRowView");

        // this.title_node.textContent = this.model.get("title")
        // this.uniques_node.style = `background: linear-gradient(90deg, #A1D8FF ${this.unique_percent}, transparent 0%);`

        var new_node // placeholder

        var loading_disclaimer = ""
        if (this.loading) {
            loading_disclaimer = `<br><br>This result is based on a <b>random sample of ${this.sample_size} rows</b>. We are still calculating the results for the full dataset in the background.`
            new_node = $(`<div class="bamboolib-glimpse-loader"><div class='bamboolib-load-spinner'></div></div>`)[0]
        } else {
            new_node = $(`<div></div>`)[0]
        }
        this.loading_node.replaceWith(new_node)
        this.loading_node = new_node

        new_node = $(`<div class="bamboolib-glimpse-column-name">${this.column_name}</div>`)[0]
        this.column_name_node.replaceWith(new_node)
        this.column_name_node = new_node

        new_node = $(`<div class="bamboolib-glimpse-dtype">${this.dtype}</div>`)[0]
        this.dtype_node.replaceWith(new_node)
        this.dtype_node = new_node

        new_node = $(`
            <div
                title="Unique values: ${this.unique_count} - ${this.unique_percent}${loading_disclaimer}"
                class="bamboolib-glimpse-unique bamboolib-tooltip"
                style="background: linear-gradient(90deg, #A1D8FF ${this.unique_percent}, transparent 0%);"
                >
            ${this.unique_count}
            </div>`)[0]
        this.uniques_node.replaceWith(new_node)
        this.maybeDestroyTooltip(this.uniques_node)
        this.uniques_node = new_node

        var missings_text = (this.missings_count == "0") ? "0" : `${this.missings_count} - ${this.missings_percent}`
        new_node = $(`
            <div
                title="Missing values: ${this.missings_count} - ${this.missings_percent}${loading_disclaimer}"
                class="bamboolib-glimpse-missing bamboolib-tooltip"
                style="
                    background: linear-gradient(90deg, #d65f5f ${this.missings_percent}, transparent 0%);
                    background-color: #5fba7d;"
                >
            ${missings_text}
            </div>`)[0]
        this.missings_node.replaceWith(new_node)
        this.maybeDestroyTooltip(this.missings_node)
        this.missings_node = new_node

        $(function () {
            // function needs to be executed after the elements have been rendered
            $(".bamboolib-tooltip").tooltip(
                {
                    content: function () {
                        // enables to render HTML within the tooltip
                        return $(this).prop('title');
                    },
                    hide: false,
                    show: false,
                    position: { my: "left top+3", at: "left bottom", collision: "flipfit" },
                    // position: { my: "center top+10", at: "center bottom", collision: "flipfit" }
                })
        })
    }

    initialize(parameters) {
        super.initialize(parameters);
        // console.log("initialize BamGlimpseRow")
        // window.flolo = this

        this.column_name_node = this.$el[0].appendChild($(`<div></div>`)[0])
        this.dtype_node = this.$el[0].appendChild($(`<div></div>`)[0])
        this.uniques_node = this.$el[0].appendChild($(`<div></div>`)[0])
        this.missings_node = this.$el[0].appendChild($(`<div></div>`)[0])
        this.loading_node = this.$el[0].appendChild($(`<div></div>`)[0])

        this.update()

        this.$el.on("click", (event) => {
            var msg = {
                'type': 'on_click',
            };
            this.send(msg);
            // console.log("BamGlimpseRow click", true)
        })
    }
}

class BamFocusBoxModel extends widgets_controls.BoxModel {
    defaults() {
        // console.log("init BamFocusBoxModel")
        return _.extend(widgets_controls.BoxModel.prototype.defaults(), {
            _model_name: 'BamFocusBoxModel',
            _view_name: 'BamFocusBoxView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
        });
    }
}

class BamFocusBoxView extends widgets_controls.BoxView {
    initialize(parameters) {
        super.initialize(parameters);
        // console.log("initialize BamFocusBox")
        // window.flolo = this

        // enable the div to be focusable in order to be able to catch the keyboard events
        this.$el[0].tabIndex = 0

        // input if we want to control the behavior of what gets selected after a tab:
        // we might create a loop, eg selecting the first element in a row again
        // we might not allow to focus past a certain element
        // https://stackoverflow.com/questions/7208161/focus-next-element-in-tab-index

        this.$el.on("keydown", (event) => {
            // console.log("FocusBox keydown", true)
            // dont propagate event to jupyter because this will interact with the editor
            event.stopPropagation()

            // // prevent the default browser action, eg scrolling on mouse up or down arrow
            // // but this also prevents the typing in the controls
            // event.preventDefault()
        })
            .on("keyup", (event) => {
                // console.log("FocusBox keyup", true)
                event.stopPropagation()
            })
            .on("keypress", (event) => {
                // console.log("FocusBox keypress", true)
                event.stopPropagation()
            })
            .on("click", (event) => {
                // console.log("FocusBox click", true)
                event.stopPropagation()
            })
    }
}


class BamMainThreadTriggerModel extends widgets_controls.BoxModel {
    defaults() {
        // console.log("init BamMainThreadTriggerModel")
        return _.extend(widgets_controls.BoxModel.prototype.defaults(), {
            _model_name: 'BamMainThreadTriggerModel',
            _view_name: 'BamMainThreadTriggerView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
        });
    }
}


class BamMainThreadTriggerView extends widgets_controls.BoxView {
    initialize(parameters) {
        super.initialize(parameters);
        // console.log("initialize BamMainThreadTrigger")
        // window.flolo = this

        this.model.on('msg:custom', this.echo_message, this);

        var message = {
            'type': 'on_init',
        }
        this.send(message);
    }

    echo_message(msg) {
        this.send(msg);
    }
}


class BamFocusPointModel extends widgets_controls.BoxModel {
    defaults() {
        // console.log("init BamFocusPointModel")
        return _.extend(widgets_controls.BoxModel.prototype.defaults(), {
            _model_name: 'BamFocusPointModel',
            _view_name: 'BamFocusPointView',

            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
        });
    }
}

class BamFocusPointView extends widgets_controls.BoxView {
    initialize(parameters) {
        super.initialize(parameters);
        // console.log("initialize BamFocusPoint")
        // window.flolo = this

        // set tabIndex so that the div becomes focusable
        this.$el[0].tabIndex = -1
        this.$el.focus()

        this.model.on('msg:custom', this.focus, this);
        // when trying focus from the Chrome Debug tools, we need to use setTimeout and then click in the main window
        // setTimeout(function(){flolo.$el.focus()}, 1000)
    }
    focus(){
        this.$el.focus()
    }
}

class BamSelectizeModel extends widgets.DOMWidgetModel {
    defaults() {
        return _.extend(widgets.DOMWidgetModel.prototype.defaults(), {
            _model_name: 'BamSelectizeModel',
            _view_name: 'BamSelectizeView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
            value: '',
            // _columns: {}
        });
    }
}

class BamSelectizeView extends widgets.DOMWidgetView {
    // code inspiration from BamAutocompleteTextView
    render() {
        this.model.on('msg:custom', this.handle_msg, this);
        this.listenTo(this.model, 'change:itemsSelectize', () => this._updateItemsSelectize(this.model.get("itemsSelectize")))
        this.listenTo(this.model, 'change:optionsSelectize', () => this.render())
        this.ignore_next_update = false
        // console.log("BamSelectizeView render")
        // window.trol = this  // debug
        $(this.$el[0]).empty()  // maybe reset
        // // type of input is handled by selectize via maxItems
        this.input = $(`<input class="bamboolib-selectize" type="text">`).appendTo(this.$el);
        // this.input = $(`<input type="text" id="lol">`).appendTo(this.$el);

        // id="lol"
        // $("#lol")[0].selectize
        // $("#lol")[0].selectize.items
        // $("#lol")[0].selectize.$control_input

        var optionsSelectize = this.model.get("optionsSelectize")
        // console.log("selectize optionsSelectize", optionsSelectize)
        var itemsSelectize = this.model.get("itemsSelectize")
        // console.log("selectize itemsSelectize", itemsSelectize)
        var maxItems = this.model.get("_max_items")
        // console.log("selectize maxItems", maxItems)
        var isDropdown = maxItems == 1
        this.isMultiSelect = !isDropdown
        // var optionsSelectizeList = [
        //     {
        //         text: 'Option One',
        //         value: 1
        //     },
        //     {
        //         text: 'Option Two',
        //         value: 2
        //     }
        // ];

        var placeholder = this.model.get("placeholder")
        placeholder += "  "  // trailing whitespace in order to prevent the bug of cutting the string

        // ENTER selects search in selectize, TAB changes to next element
        // dropdowns can also be selected (and switched away from) with TAB
        var selectOnTab = this.model.get("select_on_tab") || isDropdown
        // var selectOnTab = true

        this.userIsSearching = false

        this.input.selectize({
            create: false,  // user can add/create new items
            placeholder: placeholder,
            options: optionsSelectize,
            items: itemsSelectize,
            maxItems: maxItems,
            searchField: ['value', 'description'],
            selectOnTab: selectOnTab,
            // closeAfterSelect: true,
            render: {
                // Defines the appearance of displayed option
                option: function (data, escape) {
                    if (data.description === undefined || data.description === '') {
                        return '<div class="option">' + escape(data.value) + '</div>'
                    } else {
                        return '<div class="option">' +
                            '<span class="value">' + escape(data.value) + '</span>' +
                            '<span class="description">' + escape(data.description) + '</span>' +
                            '</div>';
                    }
                },
                // Defines the appearance of a selected item
                item: function (data, escape) {
                    return '<div class="item">' + escape(data.value) + '</div>';
                }
            },
            onChange: (value) => { this.onChange(value) },
            onFocus: () => {
                if (isDropdown) {
                    var _is_soft_value = this.model.get("_is_soft_value")
                    if (_is_soft_value) {
                        // clear selection so that the users can just type
                        this.selectize.clear(true)
                        this.model.set('_is_soft_value', false)
                        this.touch()
                    }
                }
            },
            onType: (str) => {
                // console.log("onType", str)
                if (str == '') {
                    this.userIsSearching = false
                } else {
                    this.userIsSearching = true
                }
                this.updateSelectOnTab()
            },
            onItemAdd: () => {
                // console.log("onItemAdd")
                this.userIsSearching = false
                this.updateSelectOnTab()
            }
            // onItemRemove: () => {
            //     console.log("onItemRemove")
            //     if(isMultiSelect){
            //         this.selectize.close()
            //     }
            // }
        });
        this.selectize = this.input[0].selectize
        this.preventEventPropagation()

        var enabled = this.model.get("enabled")
        if (!enabled) {
            this.selectize.disable()
        }

        var focus_after_init = this.model.get("focus_after_init")
        var _did_focus_after_init = this.model.get("_did_focus_after_init")
        if (focus_after_init && !_did_focus_after_init) {
            this.selectize.focus()
            this.model.set('_did_focus_after_init', true)
            this.touch()
        }
    }

    updateSelectOnTab() {
        if (this.model.get("select_on_tab")) {
            // dont update
            return true
        }
        else if (this.isMultiSelect) {
            var selectOnTab = false
            if (this.userIsSearching) {
                selectOnTab = true
            }
            this.selectize.settings.selectOnTab = selectOnTab
            // console.log("updated selectonTab to", selectOnTab)
        }
    }

    handle_msg(msg) {
        // console.log("handle_msg", msg)
        if (msg.type === 'focus') {
            this.selectize.focus()
        }
        if (msg.type === 'rerender') {
            this.render()
        }
    }

    preventEventPropagation() {
        // there are also keypress and keyup events, but CodeMirror is currently listening to keydown
        this.selectize.$control_input[0].addEventListener("keydown", (event) => {
            // console.log("keydown", event)
            // dont propagate enter or tab event to CodeMirror because this will focus the editor
            // and this might lead to a scroll event by CodeMirror
            event.stopPropagation();
        });
    }

    onChange(value) {
        this.ignore_next_update = true

        var itemsSelectize = this.selectize.items
        // console.log("selectize onChange", itemsSelectize)

        var message = {
            'type': 'on_change',
            'itemsSelectize': itemsSelectize
        }
        this.send(message);
    }

    _updateItemsSelectize(itemsSelectize) {
        if (this.ignore_next_update) {
            this.ignore_next_update = false
        } else {
            // console.log("_updateItemsSelectize", itemsSelectize)
            this.selectize.clear(true)
            this.selectize.setValue(itemsSelectize, true)
            // for(let item of itemsSelectize){
            //     console.log("for item", item)
            //     this.selectize.addItem(item, true)
            // }
        }
    }
}


// just inheriting from widgets_controls.TextModel did not work because the focus on the input did not work
class BamTextModel extends widgets.DOMWidgetModel {
    defaults() {
        // console.log("init BamTextModel")
        return _.extend(widgets.DOMWidgetModel.prototype.defaults(), {
            _model_name: 'BamTextModel',
            _view_name: 'BamTextView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
            value: '',
        });
    }
}

class BamTextView extends widgets.DOMWidgetView {
    update_value(value) {
        this.input[0].value = value
    }

    update_model() {
        // console.log("update_model", this.input[0].value)
        this.model.set('value', this.input[0].value)
        this.touch()
        // var msg = {
        //     'type': 'on_value_change',
        //     'value': this.input[0].value
        // };
        // this.send(msg);
        // console.log("msg", msg)
    }

    render() {
        // window.flolo = this
        // console.log("render")

        this.listenTo(this.model, 'change:value', (model, value, options) => {
            // console.log("change:value", value)
            this.update_value(value)
        });

        this.input = $(`<input type="text" style="width: 100%; line-height: inherit; font-size: inherit; box-sizing: inherit;" autocomplete="new-password"></input>`).appendTo(this.$el);
        this.update_value(this.model.get("value"))
        this.input[0].placeholder = this.model.get("placeholder")
        // this.input[0].onchange = () => {
        //     console.log("input onchange")
        //     this.update_model()
        // }


        this.input
            .ready(() => {
                var focus_after_init = this.model.get("focus_after_init")
                var _did_focus_after_init = this.model.get("_did_focus_after_init")
                if (focus_after_init && !_did_focus_after_init) {
                    // console.log("try to focus")
                    // console.log("current focus:", this.input.is(":focus"))
                    this.input[0].focus()
                    this.model.set('_did_focus_after_init', true)
                    this.touch()
                }
            })
            .enterKey(() => {
                // console.log("enterkey")
                var msg = {
                    'type': 'on_press_enter',
                };
                this.send(msg);
            })
            .on("input", (event) => {
                // before, we had the problem, that we did the update but the input.value was not set yet
                // input.value is only guaranteed to be set after on:input
                // thus, we should not use keypress or others. They all have problems
                // also, the order is: keydown, keypress, keyup
                // https://stackoverflow.com/questions/8795283/jquery-get-input-value-after-keypress
                // console.log("input")
                this.update_model()
            })
    }
}


class ButtonModel extends widgets_controls.ButtonModel {
    defaults() {
        return _.extend(widgets_controls.ButtonModel.prototype.defaults(), {
            _model_name: 'ButtonModel',
            _view_name: 'ButtonView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
        });
    }
}

class ButtonView extends widgets_controls.ButtonView {
    render() {
        // this code is copied from ipywidgets.Button AND we changed the css class
        this.listenTo(this.model, 'change:button_style', this.update_button_style);
        this.el.classList.add('bamboolib-base-button');
        this.set_button_style();
        this.update(); // Set defaults.
    }
}


class BamCopyButtonModel extends ButtonModel {
    defaults() {
        // console.log("init BamCopyButtonModel")
        return _.extend(ButtonModel.prototype.defaults(), {
            _model_name: 'BamCopyButtonModel',
            _view_name: 'BamCopyButtonView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
            copy_string: '',
        });
    }
}

class BamCopyButtonView extends ButtonView {
    // render() {
    //     // window.flolo = this
    //     console.log("render")
    //     super.render()
    // }

    update() {
        // ATTENTION: the update method is called after the first render
        // AND after all model changes
        // during the update method of widgets.ButtonView the button gets restored
        // thus, the textarea element gets deleted and we need to restore it
        // This is also the reason, why the textarea element is created here (again and again)
        // and not only once in render

        // console.log("update")
        super.update()

        this.textarea = $(`<textarea tabindex="-1"></textarea>`).appendTo(this.$el)
        this.textarea[0].value = this.model.get("copy_string")
    }

    _handle_click(event) {
        // console.log("BamCopy: _handle_click :))", this)

        this.textarea[0].select()
        document.execCommand("copy")

        super._handle_click(event)
    }
}


class SVGButtonModel extends ButtonModel {
    defaults() {
        // console.log("init BamCopyButtonModel")
        return _.extend(ButtonModel.prototype.defaults(), {
            _model_name: 'SVGButtonModel',
            _view_name: 'SVGButtonView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
            path: '',
            svg_size: '12',
            view_box: '',
        });
    }
}

class SVGButtonView extends ButtonView {
    render() {
        var xmlns = "http://www.w3.org/2000/svg";
        var svg = document.createElementNS(xmlns, "svg");
        svg.setAttributeNS(null, "viewBox", this.model.get("view_box"));
        svg.setAttributeNS(null, "width", this.model.get("svg_size"));
        svg.innerHTML = '<path d="' + this.model.get("path") + '"></path>';
        this.el.appendChild(svg);
        this.update(); // Set defaults.
    }
}

class BamAutocompleteTextModel extends widgets.DOMWidgetModel {
    defaults() {
        return _.extend(widgets.DOMWidgetModel.prototype.defaults(), {
            _model_name: 'BamAutocompleteTextModel',
            _view_name: 'BamAutocompleteTextView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
            value: ''
            // _columns: {}
        });
    }
}

class BamAutocompleteTextView extends widgets.DOMWidgetView {

    handle_msg(msg) {
        // console.log("handle_msg", msg)
        if (msg.type === 'completion') {
            // console.log("msg.completions", msg.completions)
            if (typeof this.response !== 'undefined') {
                // console.log("this.response", this.response)
                this.response(msg.completions)
            } else {
                // console.log("this.response is still undefined")
            }
        }
    }

    update_value() {
        this.input[0].value = this.model.get("value")
    }

    update_model() {
        // console.log("update_model", this.input[0].value)
        this.model.set('value', this.input[0].value);
        this.touch()
    }

    cursor_is_at_single_position(element) {
        return element.selectionStart == element.selectionEnd
    }

    maybe_complete(element, start_string, end_string) {
        if (this.cursor_is_at_single_position(element)) {
            var cursor_position = element.selectionEnd
            var before_cursor = this.old_input_state["text_before_cursor"]
            var after_cursor = this.old_input_state["text_after_cursor"]

            if (element.value === before_cursor + start_string + after_cursor) {
                element.value = before_cursor + start_string + end_string + after_cursor
                element.setSelectionRange(cursor_position, cursor_position);
                // console.log(`did complete ${start_string} with ${end_string}`)
            }
        }
    }

    save_input_state(element) {
        var cursor_position = element.selectionEnd
        var before_cursor = element.value.substring(0, cursor_position)
        var after_cursor = element.value.substring(cursor_position)
        this.old_input_state = {
            "text_before_cursor": before_cursor,
            "text_after_cursor": after_cursor,
            // "cursor_position": 0,
        }
    }

    render() {
        // window.flolo = this  // best way to debug

        this.model.on('msg:custom', this.handle_msg, this);
        this.listenTo(this.model, 'change:value', (model, value, options) => {
            this.update_value();
        });

        // $(`<label for="bamboolib-complete">BamAutocompleteText: </label>`).appendTo(this.$el);
        this.input = $(`<textarea type="text"
                            id="bamboolib-complete"
                            class="bamboolib-input widget-textarea"
                            placeholder="${this.model.get('placeholder')}"
                            style="margin-left: 2px;
                                margin-top:${this.model.get('margin_top')}px;
                                width: 99%;
                                font-family: monospace;
                                line-height: inherit;
                                font-size: inherit;
                                box-sizing: inherit;
                                resize: vertical;"
                            rows="${this.model.get('nrows')}"
                        ></textarea>`)
            .appendTo(this.$el);
        this.update_value()

        // // demo on custom data https://jqueryui.com/autocomplete/#custom-data
        // var completions = [
        //         {
        //             label: "df",
        //             completion: "f",
        //         },
        //         {
        //             label: "df2",
        //             completion: "f2",
        //         },
        //     ];

        this.old_input_state = {
            "text_before_cursor": "",
            "text_after_cursor": "",
            // "cursor_position": 0,
        }

        this.input
            .ready(() => {
                let focus_after_init = this.model.get("focus_after_init");
                if (focus_after_init) {
                    this.input[0].focus()
                }
            })
            .on("keypress", function (event) {
                // don't navigate away from the field on tab when selecting an item
                if (event.keyCode === $.ui.keyCode.TAB &&
                    $(this).autocomplete("instance").menu.active) {
                    event.preventDefault();
                }
            })
            .enterKey((event) => {
                event.preventDefault();
                var msg = {
                    'type': 'on_press_enter',
                };
                this.send(msg);
            })
            .on("keyup", (event) => {
                var input = this.input[0]
                this.maybe_complete(input, "(", ")")
                this.maybe_complete(input, "[", "]")
                this.maybe_complete(input, "{", "}")
                this.maybe_complete(input, "'", "'")
                this.maybe_complete(input, '"', '"')
                this.save_input_state(input)
                this.update_model()
            })
            .on("input", (event) => {
                // only triggered by user input but not when we change input.value programmatically?
                this.update_model()
            })
            .autocomplete({
                minLength: 0,
                delay: 10,  // perceived as real time
                autoFocus: true,
                source: (request, response) => {
                    var msg = {
                        'type': 'completion_request',
                        'input': request.term,
                        'cursor_position': this.input[0].selectionEnd,
                    };
                    this.send(msg);

                    this.response = response

                    // var last_word_after_space = request.term.split(" ").pop()
                    // var responses_list = $.ui.autocomplete.filter(
                    //     completions, last_word_after_space
                    // )
                    // // this is a string list of potential autocompletions

                    // response(completions)
                },
                focus: function () {
                    // prevent value inserted on focus
                    return false;
                },
                select: (event, ui) => {
                    // console.log("event, ui", [event, ui])
                    var input = this.input[0]
                    var old_input = input.value
                    var old_cursor_position = input.selectionEnd

                    var before_cursor = old_input.substring(0, old_cursor_position)
                    var after_cursor = old_input.substring(old_cursor_position)

                    var completion = ui.item.completion

                    var new_value = before_cursor + completion + after_cursor
                    this.input[0].value = new_value
                    var new_cursor_position = old_cursor_position + completion.length
                    input.setSelectionRange(new_cursor_position, new_cursor_position)

                    this.update_model()
                    this.save_input_state(input)

                    event.preventDefault()

                    // console.log("this.value", this.value)
                    // // shorthand for event.target.value = "loool"

                    // // data attributes for maybe saving state?
                    // // https://developer.mozilla.org/en-US/docs/Learn/HTML/Howto/Use_data_attributes
                    // // this.dataset.last_search = "whoo"

                    // // change cursor position/selection in the target
                    // // this.setSelectionRange(2, 4);

                    return false;
                }
            });
    }
}

// Model for the qgrid widget
class BamGridModel extends widgets.DOMWidgetModel {
    defaults() {
        return _.extend(widgets.DOMWidgetModel.prototype.defaults(), {
            _model_name: 'BamGridModel',
            _view_name: 'BamGridView',
            _model_module: 'bamboolib',
            _view_module: 'bamboolib',
            // Widget version infos in JS are not needed.
            // Widget version in python only needs to reference to JS package version
            // _model_module_version: '1.1.0',
            // _view_module_version: '1.1.0',
            _df_json: '',
            _columns: {}
        });
    }
}


// View for the qgrid widget
class BamGridView extends widgets.DOMWidgetView {
    render() {
        // window.flolo = this  // best way to debug
        // console.log("[nbextensions/bamboolib] BamGrid render")

        // subscribe to incoming messages from the BamGridWidget
        this.model.on('msg:custom', this.handle_msg, this);
        this.initialize_qgrid();

        this.initializeTextSelectionListener()

        // window.flolo = this  // debug
        // this.appendCodeToParentCell("hello")
        // console.log("this.$el", this.$el)
        this.send({ 'type': 'bam_grid_rendered' });
    }

    initializeTextSelectionListener() {
        // https://stackoverflow.com/questions/5669448/get-selected-texts-html-in-div
        // https://developer.mozilla.org/en-US/docs/Web/API/Selection
        // https://api.jquery.com/mouseup/

        $(this.$el).bind("mouseup", (event) => {
            // console.log("mouseup in BamGridView", event)
            var target_is_slick_cell = event.target.classList[0] === "slick-cell"

            if (target_is_slick_cell) {
                var selection = document.getSelection()
                var in_slick_cell = selection.focusNode.parentNode.classList[0] === "slick-cell"
                var single_cell = selection.anchorNode == selection.focusNode
                var is_range = !selection.isCollapsed

                if (in_slick_cell && is_range && single_cell) {
                    var column_index = event.target.classList[1].substring(1)
                    var selected_text = selection.toString()
                    // var full_node_content = selection.focusNode.parentNode.innerHTML

                    var msg = {
                        'type': 'bam_cell_text_selection',
                        'column_index': column_index,
                        'selected_text': selected_text,
                    };
                    this.send(msg);
                }
            }
        });
    }

    updateLiveCodeExport(code_export, initial_user_code) {
        // initially, we need to store the initial_user_code
        // js cannot savely store the initial user code because the widget might
        // be rerendered eg during toggle of show static UI or bamboolib UI
        // thus, we need to save it in the python df_manager state

        function maybeAdjustInitialUserCode(code) {
            code = code.split("\n")

            if (code.length == 1) {
                // only the user symbol is in the Jupyter cell
                return ""
            }

            let last_line = code[code.length - 1]
            let is_comment_line = last_line.trim().startsWith("#")
            let contains_semicolon = last_line.search(";") != -1

            if (is_comment_line || contains_semicolon) {
                // pass
            } else {
                code.pop();
            }
            return code.join("\n") + "\n";
        }

        // this is the first code export because the cell cannot be empty
        if (initial_user_code == "") {
            initial_user_code = this.getCodeOfParentCell()
            var msg = {
                'type': 'bam_initial_user_code',
                'initial_user_code': initial_user_code,
            }
            this.send(msg)
        }

        var new_cell_value = ""
        if (code_export == "") {
            new_cell_value = initial_user_code
        } else {
            let adjusted_initial_user_code = maybeAdjustInitialUserCode(initial_user_code)
            new_cell_value = adjusted_initial_user_code + code_export
        }
        this.setCodeOfParentCell(new_cell_value)
    }

    setCodeOfParentCell(new_code) {
        code_export.CodeExporter.setValue(this.$el, new_code)
    }

    getCodeOfParentCell() {
        var old_code = code_export.CodeExporter.getValue(this.$el)
        return old_code
    }

    /**
     * Main entry point for drawing the widget,
     * including toolbar buttons if necessary.
     */
    initialize_qgrid() {
        // console.log("[nbextensions/bamboolib] BamGrid initialize grid")
        this.$el.empty();
        if (!this.$el.hasClass('q-grid-container')) {
            this.$el.addClass('q-grid-container');
        }
        this.initialize_toolbar();
        this.initialize_slick_grid();
    }

    initialize_toolbar() {
        if (!this.model.get('show_toolbar')) {
            this.$el.removeClass('show-toolbar');
        } else {
            this.$el.addClass('show-toolbar');
        }

        if (this.toolbar) {
            return;
        }

        this.toolbar = $("<div class='q-grid-toolbar'>").appendTo(this.$el);

        let append_btn = (btn_info) => {
            return $(`
        <button
        class='btn btn-default'
        data-loading-text='${btn_info.loading_text}'
        data-event-type='${btn_info.event_type}'
        data-btn-text='${btn_info.text}'>
            ${btn_info.text}
        </button>
      `).appendTo(this.toolbar);
        };

        append_btn({
            loading_text: 'Adding...',
            event_type: 'add_row',
            text: 'Add Row'
        });

        append_btn({
            loading_text: 'Removing...',
            event_type: 'remove_row',
            text: 'Remove Row'
        });

        this.buttons = this.toolbar.find('.btn');
        this.buttons.attr('title',
            'Not available while there is an active filter');
        this.buttons.tooltip();
        this.buttons.tooltip({
            show: { delay: 300 }
        });
        this.buttons.tooltip({
            hide: { delay: 100, 'duration': 0 }
        });
        this.buttons.tooltip('disable');

        this.full_screen_btn = null;
        if (dialog) {
            this.full_screen_modal = $('body').find('.qgrid-modal');
            if (this.full_screen_modal.length == 0) {
                this.full_screen_modal = $(`
          <div class="modal qgrid-modal">
            <div class="modal-dialog">
              <div class="modal-content">
                <div class="modal-body"></div>
              </div>
            </div>
          </div>
        `).appendTo($('body'));
            }
            this.full_screen_btn = $(`
        <button
          class='btn btn-default fa fa-arrows-alt full-screen-btn'/>
      `).appendTo(this.toolbar);
            this.close_modal_btn = $(`
        <button
          class='btn btn-default fa fa-times close-modal-btn'
          data-dismiss="modal"/>
      `).appendTo(this.toolbar);

        }
        this.bind_toolbar_events();
    }

    bind_toolbar_events() {
        this.buttons.off('click');
        this.buttons.click((e) => {
            let clicked = $(e.target);
            if (clicked.hasClass('disabled')) {
                return;
            }
            if (this.in_progress_btn) {
                alert(`
          Adding/removing row is not available yet because the
          previous operation is still in progress.
        `);
            }
            this.in_progress_btn = clicked;
            clicked.text(clicked.attr('data-loading-text'));
            clicked.addClass('disabled');
            this.send({ 'type': clicked.attr('data-event-type') });
        });
        if (!this.full_screen_btn) {
            return;
        }
        this.full_screen_btn.off('click');
        this.full_screen_btn.click((e) => {
            this.$el_wrapper = this.$el.parent();
            this.$el_wrapper.height(this.$el_wrapper.height());
            this.$el.detach();
            var modal_options = {
                body: this.$el[0],
                show: false
            };
            if (IPython && IPython.keyboard_manager) {
                modal_options.keyboard_manager = IPython.keyboard_manager;
            }
            var qgrid_modal = dialog.modal(modal_options);

            qgrid_modal.removeClass('fade');
            qgrid_modal.addClass('qgrid-modal');
            qgrid_modal.on('shown.bs.modal', (e) => {
                this.slick_grid.resizeCanvas();
            });
            qgrid_modal.on('hidden.bs.modal', (e) => {
                this.$el.detach();
                this.$el_wrapper.height('auto');
                this.$el_wrapper.append(this.$el);
                this.update_size();
                this.slick_grid.bindAllEvents();
                this.bind_toolbar_events();
            });
            qgrid_modal.modal('show');
        });
    }

    /**
     * Create the grid portion of the widget, which is an instance of
     * SlickGrid, plus automatically created filter controls based on the
     * type of data in the columns of the DataFrame provided by the user.
     */
    initialize_slick_grid() {

        if (!this.grid_elem) {
            this.grid_elem = $("<div class='q-grid'>").appendTo(this.$el);
        }

        // create the table
        var df_json = JSON.parse(this.model.get('_df_json'));
        var columns = this.model.get('_columns');
        this.data_view = this.create_data_view(df_json.data);
        this.grid_options = this.model.get('grid_options');
        this.index_col_name = this.model.get("_index_col_name");
        this.row_styles = this.model.get("_row_styles");

        this.columns = [];
        this.index_columns = [];
        this.filters = {};
        this.filter_list = [];
        this.date_formats = {};
        this.last_vp = null;
        this.sort_in_progress = false;
        this.sort_indicator = null;
        this.resizing_column = false;
        this.ignore_selection_changed = false;
        this.vp_response_expected = false;
        this.next_viewport_msg = null;

        var number_type_info = {
            // filter: slider_filter.SliderFilter,
            // validator: editors.validateNumber,
            formatter: this.format_number
        };

        var self = this;

        this.type_infos = {
            integer: Object.assign(
                // { editor: Slick.Editors.Integer },
                number_type_info
            ),
            number: Object.assign(
                // { editor: Slick.Editors.Float },
                number_type_info
            ),
            string: {
                // filter: text_filter.TextFilter,
                // editor: Slick.Editors.Text,
                // formatter: this.format_string
            },
            datetime: {
                // filter: date_filter.DateFilter,
                // editor: class DateEditor extends Slick.Editors.Date {
                //     constructor(args) {
                //         super(args);

                //         this.loadValue = (item) => {
                //             this.date_value = item[args.column.field];
                //             var formatted_val = self.format_date(
                //                 this.date_value, args.column.field
                //             );
                //             this.input = $(args.container).find('.editor-text');
                //             this.input.val(formatted_val);
                //             this.input[0].defaultValue = formatted_val;
                //             this.input.select();
                //             this.input.on("keydown.nav", function (e) {
                //                 if (e.keyCode === $.ui.keyCode.LEFT || e.keyCode === $.ui.keyCode.RIGHT) {
                //                     e.stopImmediatePropagation();
                //                 }
                //             });
                //         };

                //         this.isValueChanged = () => {
                //             return this.input.val() != this.date_value;
                //         };

                //         this.serializeValue = () => {
                //             if (this.input.val() === "") {
                //                 return null;
                //             }
                //             var parsed_date = moment.parseZone(
                //                 this.input.val(), "YYYY-MM-DD HH:mm:ss.SSS"
                //             );
                //             return parsed_date.format("YYYY-MM-DDTHH:mm:ss.SSSZ");
                //         };
                //     }
                // },
                // formatter: (row, cell, value, columnDef, dataContext) => {
                //     if (value === null) {
                //         return "NaT";
                //     }
                //     return this.format_date(value, columnDef.name);
                // }
            },
            any: {
                // filter: text_filter.TextFilter,
                // editor: editors.SelectEditor,
                // formatter: this.format_string
            },
            interval: {
                // formatter: this.format_string
            },
            boolean: {
                // filter: boolean_filter.BooleanFilter,
                // editor: Slick.Editors.Checkbox,
                // formatter: (row, cell, value, columngDef, dataContext) => {
                //     return value ? `<span class="fa fa-check"/>` : "";
                // }
            }
        };

        $.datepicker.setDefaults({
            gotoCurrent: true,
            dateFormat: $.datepicker.ISO_8601,
            constrainInput: false,
            "prevText": "",
            "nextText": ""
        });

        var sorted_columns = Object.values(columns).sort(
            (a, b) => a.position - b.position
        );

        for (let cur_column of sorted_columns) {
            // console.log("[nbextensions/bamboolib] BamGrid render column " + cur_column.name)
            if (cur_column.name == this.index_col_name) {
                continue;
            }

            var type_info = this.type_infos[cur_column.type] || {};

            var slick_column = cur_column;

            Object.assign(slick_column, type_info);

            if (cur_column.type == 'any') {
                slick_column.editorOptions = {
                    options: cur_column.constraints.enum
                };
            }

            if (slick_column.filter) {
                // initializes the filter for the column
                // implicit initializes the filter object which was set above
                // eg if slick_column.filter = text_filter.TextFilter
                // then a new text_filter.TextFilter is initialized
                var cur_filter = new slick_column.filter(
                    slick_column.field,
                    cur_column.type,
                    this
                );
                this.filters[slick_column.id] = cur_filter;
                this.filter_list.push(cur_filter);
            }

            if (cur_column.width == null) {
                delete slick_column.width;
            }

            if (cur_column.maxWidth == null) {
                delete slick_column.maxWidth;
            }

            // don't allow editing index columns
            if (cur_column.is_index) {
                slick_column.editor = editors.IndexEditor;

                if (cur_column.first_index) {
                    slick_column.cssClass += ' first-idx-col';
                }
                if (cur_column.last_index) {
                    slick_column.cssClass += ' last-idx-col';
                }

                slick_column.name = cur_column.index_display_text;
                slick_column.level = cur_column.level;

                if (this.grid_options.boldIndex) {
                    slick_column.cssClass += ' idx-col';
                }

                this.index_columns.push(slick_column);
                continue;
            }

            if (cur_column.editable == false) {
                slick_column.editor = null;
            }

            this.columns.push(slick_column);
        }

        if (this.index_columns.length > 0) {
            this.columns = this.index_columns.concat(this.columns);
        }

        var row_count = 0;

        // set window.slick_grid for easy troubleshooting in the js console
        window.slick_grid = this.slick_grid = new Slick.Grid(
            this.grid_elem,
            this.data_view,
            this.columns,
            this.grid_options
        );
        this.grid_elem.data('slickgrid', this.slick_grid);

        if (this.grid_options.forceFitColumns) {
            this.grid_elem.addClass('force-fit-columns');
        }

        if (this.grid_options.highlightSelectedCell) {
            this.grid_elem.addClass('highlight-selected-cell');
        }

        // compare to false since we still want to show row
        // selection if this option is excluded entirely
        if (this.grid_options.highlightSelectedRow != false) {
            this.grid_elem.addClass('highlight-selected-row');
        }

        setTimeout(() => {
            this.slick_grid.init();
            this.update_size();
        }, 1);

        this.slick_grid.setSelectionModel(new Slick.RowSelectionModel());
        this.slick_grid.setCellCssStyles("grouping", this.row_styles);
        this.slick_grid.render();

        this.update_size();

        var render_header_cell = (e, args) => {
            // var cur_filter = this.filters[args.column.id];
            // if (cur_filter) {
            //     cur_filter.render_filter_button($(args.node), this.slick_grid);
            // }
            dtype_button.render_dtype_button(args.column, $(args.node), this)

            var header_click = (event) => {
                // console.log("header_click", event)
                // console.log("header_click args", args.column)
                if (this.resizing_column) {
                    return;
                }

                if ('is_index' in args.column) {
                    // dont notify when clicking on index button
                } else {
                    var msg = {
                        'type': 'bam_rename_column',
                        'field': args.column.field
                    };
                    this.send(msg);
                }
            }
            $(args.node).click(header_click)
            // console.log("render_header_cell", $(args.node))
        };
        this.slick_grid.onHeaderCellRendered.subscribe(render_header_cell);

        // if (this.grid_options.filterable != false) {
        //     this.slick_grid.onHeaderCellRendered.subscribe(render_header_cell);
        // }


        // Force the grid to re-render the column headers so the
        // onHeaderCellRendered event is triggered.
        this.slick_grid.setColumns(this.slick_grid.getColumns());

        $(window).resize(() => {
            this.slick_grid.resizeCanvas();
        });

        this.slick_grid.setSortColumns([]);


        // needed later:
        this.grid_header = this.$el.find('.slick-header-columns');


        // var handle_header_click = (e) => {
        //     if (this.resizing_column) {
        //         return;
        //     }

        //     if (this.sort_in_progress) {
        //         return;
        //     }

        //     var col_header = $(e.target).closest(".slick-header-column");
        //     if (!col_header.length) {
        //         return;
        //     }


        //     // console.log("column_header_click", col_header.data("column"))
        //     if ('is_index' in col_header.data("column")) {
        //         // dont notify when clicking on index button
        //     } else {
        //         var msg = {
        //             'type': 'bam_rename_column',
        //             'field': col_header.data("column").field
        //         };
        //         this.send(msg);
        //     }


        //     // // old logic for sorting the column
        //     // var column = col_header.data("column");
        //     // if (column.sortable == false) {
        //     //     return;
        //     // }

        //     // this.sort_in_progress = true;

        //     // if (this.sorted_column == column) {
        //     //     this.sort_ascending = !this.sort_ascending;
        //     // } else {
        //     //     this.sorted_column = column;
        //     //     if ('defaultSortAsc' in column) {
        //     //         this.sort_ascending = column.defaultSortAsc;
        //     //     } else {
        //     //         this.sort_ascending = true;
        //     //     }
        //     // }

        //     // var all_classes = 'fa-sort-asc fa-sort-desc fa fa-spin fa-spinner';
        //     // var clicked_column_sort_indicator = col_header.find('.slick-sort-indicator');
        //     // if (clicked_column_sort_indicator.length == 0) {
        //     //     clicked_column_sort_indicator =
        //     //         $("<span class='slick-sort-indicator'/>").appendTo(col_header);
        //     // }

        //     // this.sort_indicator = clicked_column_sort_indicator;
        //     // this.sort_indicator.removeClass(all_classes);
        //     // this.grid_elem.find('.slick-sort-indicator').removeClass(all_classes);
        //     // this.sort_indicator.addClass(`fa fa-spinner fa-spin`);
        //     // var msg = {
        //     //     'type': 'change_sort',
        //     //     'sort_field': this.sorted_column.field,
        //     //     'sort_ascending': this.sort_ascending
        //     // };
        //     // this.send(msg);
        // };

        // this.grid_header.click(handle_header_click)

        // // old logic for sorting the column
        // if (this.grid_options.sortable != false) {
        //     this.grid_header.click(handle_header_click)
        // }

        this.slick_grid.onViewportChanged.subscribe((e) => {
            if (this.viewport_timeout) {
                clearTimeout(this.viewport_timeout);
            }
            this.viewport_timeout = setTimeout(() => {
                this.last_vp = this.slick_grid.getViewport();
                var cur_range = this.model.get('_viewport_range');

                if (this.last_vp.top != cur_range[0] || this.last_vp.bottom != cur_range[1]) {
                    var msg = {
                        'type': 'change_viewport',
                        'top': this.last_vp.top,
                        'bottom': this.last_vp.bottom
                    };
                    if (this.vp_response_expected) {
                        this.next_viewport_msg = msg
                    } else {
                        this.vp_response_expected = true;
                        this.send(msg);
                    }
                }
                this.viewport_timeout = null;
            }, 100);
        });

        // set up callbacks
        let editable_rows = this.model.get('_editable_rows');
        if (editable_rows && Object.keys(editable_rows).length > 0) {
            this.slick_grid.onBeforeEditCell.subscribe((e, args) => {
                editable_rows = this.model.get('_editable_rows');
                return editable_rows[args.item[this.index_col_name]]
            });
        }

        this.slick_grid.onCellChange.subscribe((e, args) => {
            var column = this.columns[args.cell].name;
            var data_item = this.slick_grid.getDataItem(args.row);
            var msg = {
                'row_index': data_item.row_index, 'column': column,
                'unfiltered_index': data_item[this.index_col_name],
                'value': args.item[column], 'type': 'edit_cell'
            };
            this.send(msg);
        });

        this.slick_grid.onSelectedRowsChanged.subscribe((e, args) => {
            if (!this.ignore_selection_changed) {
                var msg = { 'rows': args.rows, 'type': 'change_selection' };
                this.send(msg);
            }
        });

        setTimeout(() => {
            this.$el.closest('.output_wrapper')
                .find('.out_prompt_overlay,.output_collapsed').click(() => {
                    setTimeout(() => {
                        this.slick_grid.resizeCanvas();
                    }, 1);
                });

            this.resize_handles = this.grid_header.find('.slick-resizable-handle');
            this.resize_handles.mousedown((e) => {
                this.resizing_column = true;
            });
            $(document).mouseup(() => {
                // wait for the column header click handler to run before
                // setting the resizing_column flag back to false
                setTimeout(() => {
                    this.resizing_column = false;
                }, 1);
            });
        }, 1);
    }

    processPhosphorMessage(msg) {
        super.processPhosphorMessage(msg)
        switch (msg.type) {
            case 'resize':
            case 'after-show':
                if (this.slick_grid) {
                    this.slick_grid.resizeCanvas();
                }
                break;
        }
    }

    has_active_filter() {
        for (var i = 0; i < this.filter_list.length; i++) {
            var cur_filter = this.filter_list[i];
            if (cur_filter.is_active()) {
                return true;
            }
        }
        return false;
    }

    /**
     * Main entry point for drawing the widget,
     * including toolbar buttons if necessary.
     */
    create_data_view(df) {
        let df_range = this.df_range = this.model.get("_df_range");
        let df_length = this.df_length = this.model.get("_row_count");
        return {
            getLength: () => {
                return df_length;
            },
            getItem: (i) => {
                if (i >= df_range[0] && i < df_range[1]) {
                    var row = df[i - df_range[0]] || {};
                    row.row_index = i;
                    return row;
                } else {
                    return { row_index: i };
                }
            }
        };
    }

    set_data_view(data_view) {
        this.data_view = data_view;
        this.slick_grid.setData(data_view);
    }

    format_date(date_string, col_name) {
        if (!date_string) {
            return "";
        }
        var parsed_date = moment.parseZone(date_string, "YYYY-MM-DDTHH:mm:ss.SSSZ");
        var date_format = null;
        if (parsed_date.millisecond() != 0) {
            date_format = `YYYY-MM-DD HH:mm:ss.SSS`;
        } else if (parsed_date.second() != 0) {
            date_format = "YYYY-MM-DD HH:mm:ss";
        } else if (parsed_date.hour() != 0 || parsed_date.minute() != 0) {
            date_format = "YYYY-MM-DD HH:mm";
        } else {
            date_format = "YYYY-MM-DD";
        }

        if (col_name in this.date_formats) {
            var old_format = this.date_formats[col_name];
            if (date_format.length > old_format.length) {
                this.date_formats[col_name] = date_format;
                setTimeout(() => {
                    this.slick_grid.invalidateAllRows();
                    this.slick_grid.render();
                }, 1);
            } else {
                date_format = this.date_formats[col_name];
            }
        } else {
            this.date_formats[col_name] = date_format;
        }

        return parsed_date.format(date_format);
    }

    format_string(row, cell, value, columnDef, dataContext) {
        return value;
    }

    format_number(row, cell, value, columnDef, dataContext) {
        if (value === null) {
            return 'NaN';
        }
        return value;
    }

    /**
     * Handle messages from the BamGridWidget.
     */
    handle_msg(msg) {
        if (msg.type == "bam_live_code_export") {
            this.updateLiveCodeExport(msg.code_export, msg.initial_user_code)
        } else if (msg.type == "bam_refresh_grid") {
            this.slick_grid.resizeCanvas()
        } else if (msg.type === 'draw_table') {
            this.initialize_slick_grid();
        } else if (msg.type == 'show_error') {
            alert(msg.error_msg);
            if (msg.triggered_by == 'add_row' ||
                msg.triggered_by == 'remove_row') {
                this.reset_in_progress_button();
            }
        } else if (msg.type == 'update_data_view') {
            if (this.buttons) {
                if (this.has_active_filter()) {
                    this.buttons.addClass('disabled');
                    this.buttons.tooltip('enable');
                } else if (this.buttons.hasClass('disabled')) {
                    this.buttons.removeClass('disabled');
                    this.buttons.tooltip('disable');
                }
            }
            if (this.update_timeout) {
                clearTimeout(this.update_timeout);
            }
            this.update_timeout = setTimeout(() => {
                var df_json = JSON.parse(this.model.get('_df_json'));
                this.row_styles = this.model.get("_row_styles");
                this.multi_index = this.model.get("_multi_index");
                var data_view = this.create_data_view(df_json.data);

                if (msg.triggered_by === 'change_viewport') {
                    if (this.next_viewport_msg) {
                        this.send(this.next_viewport_msg);
                        this.next_viewport_msg = null;
                        return;
                    } else {
                        this.vp_response_expected = false;
                    }
                }

                if (msg.triggered_by == 'change_sort' && this.sort_indicator) {
                    var asc = this.model.get('_sort_ascending');
                    this.sort_indicator.removeClass(
                        'fa-spinner fa-spin fa-sort-asc fa-sort-desc'
                    );
                    var fa_class = asc ? 'fa-sort-asc' : 'fa-sort-desc';
                    this.sort_indicator.addClass(fa_class);
                    this.sort_in_progress = false;
                }

                let top_row = null;
                if (msg.triggered_by === 'remove_row') {
                    top_row = this.slick_grid.getViewport().top;
                }

                this.set_data_view(data_view);

                var skip_grouping = false;
                if (this.multi_index) {
                    for (var i = 1; i < this.filter_list.length; i++) {
                        var cur_filter = this.filter_list[i];
                        if (cur_filter.is_active()) {
                            skip_grouping = true;
                        }
                    }
                }

                if (skip_grouping) {
                    this.slick_grid.removeCellCssStyles("grouping");
                } else {
                    this.slick_grid.setCellCssStyles("grouping", this.row_styles);
                }

                this.slick_grid.render();

                if ((msg.triggered_by == 'add_row' ||
                    msg.triggered_by == 'remove_row') && !this.has_active_filter()) {
                    this.update_size();
                }
                this.update_timeout = null;
                this.reset_in_progress_button();
                if (top_row) {
                    this.slick_grid.scrollRowIntoView(top_row);
                } else if (msg.triggered_by === 'add_row') {
                    this.slick_grid.scrollRowIntoView(msg.scroll_to_row);
                    this.slick_grid.setSelectedRows([msg.scroll_to_row]);
                } else if (msg.triggered_by === 'change_viewport' &&
                    this.last_vp.bottom >= this.df_length) {
                    this.slick_grid.scrollRowIntoView(this.last_vp.bottom);
                }

                var selected_rows = this.slick_grid.getSelectedRows().filter((row) => {
                    return row < Math.min(this.df_length, this.df_range[1]);
                });
                this.send({
                    'rows': selected_rows,
                    'type': 'change_selection'
                });
            }, 100);
        } else if (msg.type == 'change_grid_option') {
            var opt_name = msg.option_name;
            var opt_val = msg.option_value;
            if (this.slick_grid.getOptions()[opt_name] != opt_val) {
                this.slick_grid.setOptions({ [opt_name]: opt_val });
                this.slick_grid.resizeCanvas();
            }
        } else if (msg.type == 'change_selection') {
            this.ignore_selection_changed = true;
            this.slick_grid.setSelectedRows(msg.rows);
            if (msg.rows && msg.rows.length > 0) {
                this.slick_grid.scrollRowIntoView(msg.rows[0]);
            }
            this.ignore_selection_changed = false;
        } else if (msg.type == 'change_show_toolbar') {
            this.initialize_toolbar();
        } else if (msg.col_info) {
            var filter = this.filters[msg.col_info.name];
            filter.handle_msg(msg);
        }
    }

    reset_in_progress_button() {
        if (this.in_progress_btn) {
            this.in_progress_btn.removeClass('disabled');
            this.in_progress_btn.text(
                this.in_progress_btn.attr('data-btn-text')
            );
            this.in_progress_btn = null;
        }
    }

    /**
     * Update the size of the dataframe.
     */
    update_size() {
        var row_height = this.grid_options.rowHeight;
        var min_visible = 'minVisibleRows' in this.grid_options ?
            this.grid_options.minVisibleRows : 8;
        var max_visible = 'maxVisibleRows' in this.grid_options ?
            this.grid_options.maxVisibleRows : 15;

        var min_height = row_height * min_visible;
        // add 2 to maxVisibleRows to account for the header row and padding
        var max_height = 'height' in this.grid_options ? this.grid_options.height :
            row_height * (max_visible + 2);
        var grid_height = max_height;
        var total_row_height = (this.data_view.getLength() + 1) * row_height + 1;
        if (total_row_height <= max_height) {
            grid_height = Math.max(min_height, total_row_height);
            this.grid_elem.addClass('hide-scrollbar');
        } else {
            this.grid_elem.removeClass('hide-scrollbar');
        }
        this.grid_elem.height(grid_height);
        this.slick_grid.render();
        this.slick_grid.resizeCanvas();
    }
}

module.exports = {
    BamAutocompleteTextModel: BamAutocompleteTextModel,
    BamAutocompleteTextView: BamAutocompleteTextView,
    BamClickableModel: BamClickableModel,
    BamClickableView: BamClickableView,
    BamCopyButtonModel: BamCopyButtonModel,
    BamCopyButtonView: BamCopyButtonView,
    BamGlimpseHeaderModel: BamGlimpseHeaderModel,
    BamGlimpseHeaderView: BamGlimpseHeaderView,
    BamGlimpseRowModel: BamGlimpseRowModel,
    BamGlimpseRowView: BamGlimpseRowView,
    BamGridModel: BamGridModel,
    BamGridView: BamGridView,
    BamFocusBoxModel: BamFocusBoxModel,
    BamFocusBoxView: BamFocusBoxView,
    BamMainThreadTriggerModel: BamMainThreadTriggerModel,
    BamMainThreadTriggerView: BamMainThreadTriggerView,
    BamFocusPointModel: BamFocusPointModel,
    BamFocusPointView: BamFocusPointView,
    BamSelectizeModel: BamSelectizeModel,
    BamSelectizeView: BamSelectizeView,
    BamTabModel: BamTabModel,
    BamTabView: BamTabView,
    BamTextModel: BamTextModel,
    BamTextView: BamTextView,
    ButtonModel: ButtonModel,
    ButtonView: ButtonView,
    SVGButtonModel: SVGButtonModel,
    SVGButtonView: SVGButtonView,
};
